/* 
  https://github.com/SensorsIot/DCF77-Transmitter-for-ESP32
  Modificado
  SETA43 15/10/2024
  Para que funcione con la placa ESP-12 _versión con OLED
  seta43.duckdns.org
  seta43.blogspot.com
*/

#include <ESP8266WiFi.h>
#include <Ticker.h>


#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
#define SCREEN_ADDRESS 0x3C //OLED 128x64
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);


// Wifi credentials
const char* ssid = "yourSSID";
const char* password = "yourPassword";

int impulseArray[60];
int impulseCount = 0;
int hora, minuto, segundo, diaMes, mes, ano, diaSemana;
int timeRunningContinuous = 0;

const char* ntpServer = "bg.pool.ntp.org"; // enter your closest pool or pool.ntp.org
const char* TZ_INFO    = "CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00";  // enter your time zone (https://remotemonitoringsystems.ca/time-zone-abbreviations.php)

struct tm timeinfo;
String time_data;

Ticker tickerDecisec; // TBD at 100ms

#define MOD_DCF   0 // D3
#define OUT_DCF   4 // D2
#define LED_TEST  2 // D4

#define FRECUENCIA 76190  //76200 ->77580

char VARI[80];

void setup() 
{
  Serial.begin(9600);
  delay(1000);
  
  Serial.println("INICIO DCF77");
  Wire.begin(12, 13);  //ESP8266 ESP-12E --Wire.begin(sda_pin, scl_pin) (D6,D7)
  if(!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) 
    {
      Serial.println(F("SSD1306 allocation failed"));
      for(;;); // Don't proceed, loop forever
    }

  display.clearDisplay();
  display.setTextSize(1); // Draw 2X-scale text
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println(F("DCF77 transmisor"));
  display.println(F("SETA43"));
  display.display(); 

  pinMode (LED_TEST, OUTPUT); 
  pinMode (OUT_DCF, OUTPUT); 
  pinMode (MOD_DCF, OUTPUT); 
 
  WiFi_on();
  getNTP();

  String myIP = WiFi.localIP().toString();
  myIP.toCharArray(VARI,15);
  display.println(VARI);
  
  WiFi_off();
  CodeTime();

  display.println("");
  sprintf(VARI,"%02d/%02d/20%02d ",diaMes,mes,ano); 
  display.println(VARI);
  display.display();

  if (timeinfo.tm_isdst == 0) 
        { Serial.println("DST=OFF"); } 
    else
        { Serial.println("DST=ON");}

  tickerDecisec.attach_ms(100, DcfOut); // from now on calls DcfOut every 100ms  

  analogWriteFreq(FRECUENCIA);
  analogWrite(OUT_DCF, 128); 
}

void loop() 
{

}


void CodeTime() 
{
  diaSemana = timeinfo.tm_wday;
  if (diaSemana == 0) diaSemana = 7;
  diaMes = timeinfo.tm_mday;
  mes = timeinfo.tm_mon + 1;
  ano = timeinfo.tm_year - 100;
  hora = timeinfo.tm_hour;
  minuto = timeinfo.tm_min +1; // DCF77 transmitts the next minute
  segundo = timeinfo.tm_sec; 

  if (segundo == 0 && impulseCount==9)
  {
    sprintf(VARI,"%02d:%02d   ",hora,minuto); 
    Serial.print(VARI);
  }

  int n, Tmp, TmpIn;
  int ParityCount = 0;

  //we put the first 20 bits of each minute at a logical zero value
  for (n = 0; n < 20; n++) impulseArray[n] = 1;
  
  // set DST bit
  if (timeinfo.tm_isdst == 0) {
    impulseArray[18] = 2; // CET or DST OFF
  } else {
    impulseArray[17] = 2; // CEST or DST ON
  }
  
  //bit 20 must be 1 to indicate active time
  impulseArray[20] = 2;

  //calculates the bits for the minutes
  TmpIn = Bin2Bcd(minuto);
  for (n = 21; n < 28; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  if ((ParityCount & 1) == 0)
    impulseArray[28] = 1;
  else
    impulseArray[28] = 2;

  //calculates bits for the hours
  ParityCount = 0;
  TmpIn = Bin2Bcd(hora);
  for (n = 29; n < 35; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  if ((ParityCount & 1) == 0)
    impulseArray[35] = 1;
  else
    impulseArray[35] = 2;
  ParityCount = 0;

  //calculate the bits for the actual Day of Month
  TmpIn = Bin2Bcd(diaMes);
  for (n = 36; n < 42; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  TmpIn = Bin2Bcd(diaSemana);
  for (n = 42; n < 45; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //calculates the bits for the actualMonth
  TmpIn = Bin2Bcd(mes);
  for (n = 45; n < 50; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //calculates the bits for actual year
  TmpIn = Bin2Bcd(ano);   // 2 digit year
  for (n = 50; n < 58; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //equal date
  if ((ParityCount & 1) == 0)
    impulseArray[58] = 1;
  else
    impulseArray[58] = 2;

  //last missing pulse
  impulseArray[59] = 0; // No pulse
}

int Bin2Bcd(int dato) {
  int msb, lsb;
  if (dato < 10)
    return dato;
  msb = (dato / 10) << 4;
  lsb = dato % 10;
  return msb + lsb;
}

void DcfOut() {
  switch (impulseCount++) {
    case 0:
      if (impulseArray[segundo] != 0) {
        digitalWrite(LED_TEST, LOW);
        digitalWrite(MOD_DCF, LOW); 
      }
      break;
    case 1:
      if (impulseArray[segundo] == 1) {
        digitalWrite(LED_TEST, HIGH);
        digitalWrite(MOD_DCF, HIGH); 
      }
      break;
    case 2:
        digitalWrite(LED_TEST, HIGH);
        digitalWrite(MOD_DCF, HIGH); 
      break;
    case 9:
      impulseCount = 0;
      
      if (segundo == 1 || segundo == 15 || segundo == 21  || segundo == 29 ) {
        Serial.print("-");
        time_data+="-";
      }
      if (segundo == 36  || segundo == 42 || segundo == 45  || segundo == 50 ) {
        Serial.print("-");
        time_data+="-";
      }
      if (segundo == 28  || segundo == 35  || segundo == 58 ) {
        Serial.print("P");
        time_data+="P";
      }

      if (impulseArray[segundo] == 1) {
        Serial.print("0");
        time_data+="0";
      }
      if (impulseArray[segundo] == 2) {
        Serial.print("1");
        time_data+="1";
      }

      if (segundo == 59 ) {
        Serial.println();               
        time_data="";
        timeRunningContinuous++;
        if (timeRunningContinuous > 360) ESP.restart(); // 6 hours running, then restart all over
      }
      break;
  }
  if(!getLocalTime(&timeinfo)){
    Serial.println("Error obtaining time...");
    Serial.println(F("Error obtaining time..."));
    delay(3000);
    ESP.restart();
  }
  CodeTime();
}


void getNTP() 
{
  Serial.print("GetNTP ->");
  Serial.print(F("Syncing NTP..."));
  int i = 0;
  do {
    i++;
    if (i > 40) {
      ESP.restart();
    }
    configTime(0, 0, ntpServer);
    setenv("TZ", TZ_INFO, 1);
    delay(500);
  } while (!getLocalTime(&timeinfo));
  Serial.println("Ok");
  delay(500);
}

//________________________________________________
void WiFi_on() 
{
  byte mac[6];
  
  Serial.println("Connecting to: "+String(ssid));
  WiFi.begin(ssid, password);
  int counter = 0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    if (counter > 20) ESP.restart();
    Serial.print ( "." );
    counter++;
  }
  Serial.println();
  Serial.print(F("Connecting WiFi to:"));
  Serial.println(ssid);
  Serial.print("IP->");
  Serial.println(WiFi.localIP());
  WiFi.macAddress(mac);
  Serial.print("MAC->");
  Serial.print(mac[5],HEX);
  Serial.print(":");
  Serial.print(mac[4],HEX);
  Serial.print(":");
  Serial.print(mac[3],HEX);
  Serial.print(":");
  Serial.print(mac[2],HEX);
  Serial.print(":");
  Serial.print(mac[1],HEX);
  Serial.print(":");
  Serial.println(mac[0],HEX);
  delay(100);
}


void WiFi_off() 
{
  WiFi.disconnect(true);
  WiFi.mode(WIFI_OFF);
  Serial.println("WiFi disconnected");
  Serial.flush();
}

//-------------------------------------------
